#include <stdio.h>
#include <stdlib.h>
#include <cublas.h>
#include "cudamat_extras_kernels.cuh"
#include "cudamat.cuh"

extern "C" {

inline bool check_cublas_error() {
    cublasStatus status = cublasGetError();

    return status != CUBLAS_STATUS_SUCCESS;
}

inline bool checkCUDAError() {
    cudaError_t err = cudaGetLastError();

    if (cudaSuccess != err)
        printf("%s\n", cudaGetErrorString( err));
    return cudaSuccess != err;
}

extern const char* get_last_cuda_error() {
    cudaError_t err = cudaGetLastError();

    return cudaGetErrorString( err);
}


extern int flat_to_tiled_diag(cudamat* source, cudamat* target,
                              unsigned int img_width,
                              unsigned int num_channels, // number of colour channels
                              unsigned int tile_width,
                              unsigned int num_tiles, // number of tiles per row/column
                              unsigned int offset,
                              unsigned int num_offsets, // number of offsets per row/clumn
                              unsigned int batch_size)
{
    int channels_per_block = num_channels;

    if (tile_width * num_channels > 512)
        channels_per_block = 512 / tile_width;

    int channel_blocks = (int)ceil(((float)num_channels) / channels_per_block);

    dim3 kernelBlockGrid(num_offsets * channel_blocks, num_tiles * num_tiles, 1);
    dim3 kernelBlockDim(tile_width, channels_per_block, 1);

    kFlatToTiledDiag<<<kernelBlockGrid, kernelBlockDim>>>(source->data_device, 
                                                          target->data_device,
                                                          img_width, num_channels,
                                                          tile_width, num_tiles,
                                                          offset, num_offsets,
                                                          batch_size);
    cudaThreadSynchronize();

    if (checkCUDAError())
        return CUDA_ERROR;

    return 0;
}

extern int tiled_to_flat_diag(cudamat* source, cudamat* target, 
                              unsigned int img_width,
                              unsigned int num_channels, // number of colour channels
                              unsigned int tile_width,
                              unsigned int num_tiles, // number of tiles per row/column
                              unsigned int offset,
                              unsigned int num_offsets, // number of offsets per row/clumn
                              unsigned int batch_size)
{
    int BLOCK_WIDTH2 = 4;
    int num_blocks = (int)ceil(((float)img_width)/BLOCK_WIDTH2);

    int channels_per_block = num_channels;

    if (16 * channels_per_block > 96)
        channels_per_block = 6;

    int channel_blocks = (int)ceil(((float)num_channels) / channels_per_block);

    dim3 kernelBlockGrid(num_blocks*channel_blocks, num_blocks, 1);
    dim3 kernelBlockDim(BLOCK_WIDTH2, BLOCK_WIDTH2, channels_per_block);

    kTiledToFlatDiag<<<kernelBlockGrid, kernelBlockDim, 1024*12>>>(source->data_device,
                                                          target->data_device,
                                                          img_width, num_channels,
                                                          tile_width, num_tiles,
                                                          offset, num_offsets,
                                                          batch_size);

    cudaThreadSynchronize();

    if (checkCUDAError())
        return CUDA_ERROR;

    return 0;
}

}

