#include "cudamat_extras_kernels.cuh"
#include "float.h"


#define MAX_TILES 16
#define MAX_TILE_W 16

__global__ void kFlatToTiledDiag(float* source, float* target,
                                 unsigned int img_w,
                                 unsigned int n_chan, // number of colour channels
                                 unsigned int tile_w,
                                 unsigned int num_tiles, // number of tiles per row/column
                                 unsigned int offset,
                                 unsigned int num_offsets,
                                 unsigned int batch_size) {

    int channels_per_thread = n_chan;

    if (tile_w * n_chan > 512)
        channels_per_thread = 512 / tile_w;

    unsigned int tile_ind_x = blockIdx.y / num_tiles;
    unsigned int tile_ind_y = blockIdx.y % num_tiles;

    unsigned int offset_x = blockIdx.x % num_offsets;
    unsigned int offset_y = offset_x;
    unsigned int col_channel = (blockIdx.x / num_offsets) * channels_per_thread + threadIdx.y;

    if (col_channel < n_chan) {
        unsigned int idxSrc = col_channel * img_w * img_w;
        idxSrc += (tile_ind_x * tile_w + offset_x * offset) * img_w;
        idxSrc += tile_ind_y * tile_w + offset_y * offset;
        idxSrc += threadIdx.x;

        unsigned int tar_tile_cs = tile_w * tile_w;
        unsigned int tar_tile_s = tar_tile_cs * n_chan;
        unsigned int tar_tiling_s = num_tiles * num_tiles * tar_tile_s * batch_size;
        unsigned int idxTar = (offset_x) * (tar_tiling_s);
        idxTar += blockIdx.y * tar_tile_s;
        idxTar += col_channel * tar_tile_cs + threadIdx.x; 

        unsigned int src_item_size = img_w * img_w * n_chan;
        unsigned int tar_item_size = tar_tile_s * num_tiles * num_tiles;

        for (unsigned int i = 0; i < tile_w; i += 1) {
            for (unsigned int n = 0, t_off = 0, s_off = 0; n < batch_size; n++,
                 t_off += tar_item_size, s_off += src_item_size)
                target[idxTar + t_off] = source[idxSrc + s_off];
            idxTar += tile_w;
            idxSrc += img_w;
        }
    }
    __syncthreads();
}

__global__ void kTiledToFlatDiag(float* source, float* target,
                                 unsigned int img_width,
                                 unsigned int num_channels, // number of colour channels
                                 unsigned int tile_width,
                                 unsigned int num_tiles, // number of tiles per row/column
                                 unsigned int offset,
                                 unsigned int num_offsets,
                                 unsigned int batch_size)
{
    extern __shared__ float totals[];

    int thread_idx = threadIdx.z * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;
    int num_threads = blockDim.x * blockDim.y * blockDim.z;

    int channels_per_block = num_channels;

    if (16 * channels_per_block > 96)
        channels_per_block = 6;

    int channel_blocks = (int)ceil(((float)num_channels) / channels_per_block);

    unsigned int img_ind_x = (blockIdx.x / channel_blocks) * 4 + threadIdx.x;
    unsigned int img_ind_y = blockIdx.y * 4 + threadIdx.y;
    unsigned int col_channel = (blockIdx.x % channel_blocks) * channels_per_block + threadIdx.z;

    if (img_ind_x < img_width && img_ind_y < img_width && col_channel < num_channels) {
        for (int i = 0; i < batch_size; i++)
            totals[num_threads * i + thread_idx] = 0;

        unsigned int idxTar = col_channel * img_width * img_width + img_ind_x * img_width + img_ind_y;

        int tiling_ind_x = 0;
        int tiling_ind_y = 0;

        int tiling_size = num_channels * num_tiles * num_tiles * tile_width * tile_width * batch_size;
        int tile_size = tile_width * tile_width;
        int item_size = num_channels * tile_size * num_tiles * num_tiles;
        int img_size = img_width * img_width * num_channels;
        int cur_tiling = 0;
        int tile_ind_x = 0;
        int tile_ind_y = 0;
        int srcIdx = 0;

        for (unsigned int off_x = 0; off_x < num_offsets * offset; off_x += offset) {
            unsigned int off_y = off_x;
            tiling_ind_x = img_ind_x - off_x;
            tiling_ind_y = img_ind_y - off_y;

            if (tiling_ind_x >= 0 && tiling_ind_y >= 0 &&
                tiling_ind_x < num_tiles * tile_width && tiling_ind_y < num_tiles * tile_width) { 
                tile_ind_x = tiling_ind_x / tile_width;
                tile_ind_y = tiling_ind_y / tile_width;
                srcIdx = cur_tiling * tiling_size + (tile_ind_x * num_tiles + tile_ind_y) * num_channels * tile_size;
                srcIdx += col_channel * tile_size;
                srcIdx += (tiling_ind_x - tile_ind_x * tile_width) * tile_width;
                srcIdx += tiling_ind_y - tile_ind_y * tile_width;

                for (int i = 0, item_off = 0; i < batch_size; i++, item_off += item_size)
                    totals[num_threads * i + thread_idx] += source[srcIdx + item_off];
            }

            cur_tiling += 1;
        }
        __syncthreads();

        for (int i = 0, item_off = 0; i < batch_size; i++, item_off += img_size)
            target[idxTar + item_off] = totals[num_threads * i + thread_idx];
    }
}
