import pdb
import numpy as np
import nose
import cudamat as cm
import cudamatextras as cme


def setup():
    import gpu_lock
    id = gpu_lock.obtain_lock_id()
    cm.cuda_set_device(id)
    cm.cublas_init()

def teardown():
    cm.cublas_shutdown()

def test_flat_to_tiled_diag():
    img_w = 32 # image width
    n_chan = 64 # number of colour channels, has to be 1 for now
    tile_w = 16 # tile width, has to be a power of 2
    n_tiles = 1 # number of tiles
    off = 1 # offset
    n_off = 15 # number of offsets in each direction
    b_size = 32

    input = np.array(np.random.randn(img_w, img_w, n_chan, b_size), order='F')
    output = np.array(np.random.randn(tile_w, tile_w, n_chan, n_tiles ** 2, b_size, n_off), order='F')

    input_d = cm.CUDAMatrix(np.reshape(input.copy(), (img_w ** 2 * n_chan, b_size), order='F'))
    output_d = cm.CUDAMatrix(np.reshape(output.copy(), (tile_w**2 * n_chan, n_tiles**2 * b_size * n_off)))

    cme.flat_to_tiled_diag(input_d, output_d, n_chan, n_tiles, off, n_off)

    for off_x in range(n_off):
        off_y = off_x
        for tile_x in range(n_tiles):
            for tile_y in range(n_tiles):
                for n in range(b_size):
                    patch_x = off_x * off + tile_x * tile_w
                    patch_y = off_y * off + tile_y * tile_w

                    for c in range(n_chan):
                        sub_patch = input[patch_y:patch_y + tile_w, patch_x:patch_x + tile_w, c, n]
                        tile_ind = tile_x * n_tiles + tile_y
                        off_ind = off_x

                        output[:, :, c, tile_ind, n, off_ind] = sub_patch

    err = output_d.asarray() - np.reshape(output, output_d.shape, order='F')
    assert np.max(err) < 10**-4, "Error in cudamatextras.flat_to_tiled_diag"

def test_tiled_to_flat_diag():
    img_w = 64 # image width
    n_chan = 64 # number of colour channels, has to be 1 for now
    tile_w = 16 # tile width, has to be a power of 2
    n_tiles = 3 # number of tiles
    off = 2 # offset
    n_off = 7 # number of offsets in each direction
    b_size = 32

    input = np.array(np.random.randn(tile_w, tile_w, n_chan, n_tiles ** 2, b_size, n_off), order='F')
    #input = np.array(np.ones((tile_w, tile_w, n_chan, n_tiles ** 2, b_size, n_off ** 2)), order='F')
    output = np.array(np.zeros((img_w, img_w, n_chan, b_size)), order='F')

    input_d = cm.CUDAMatrix(np.reshape(input.copy(), (tile_w**2 * n_chan, n_tiles**2 * b_size * n_off), order='F'))
    output_d = cm.CUDAMatrix(np.reshape(output.copy(), (img_w ** 2 * n_chan, b_size), order='F'))

    cme.tiled_to_flat_diag(input_d, output_d, n_chan, n_tiles, off, n_off)

    for off_x in range(n_off):
        off_y = off_x
        for tile_x in range(n_tiles):
            for tile_y in range(n_tiles):
                for n in range(b_size):
                    patch_x = off_x * off + tile_x * tile_w
                    patch_y = off_y * off + tile_y * tile_w

                    for c in range(n_chan):
                        tile_ind = tile_x * n_tiles + tile_y
                        off_ind = off_x
    
                        output[patch_y:patch_y + tile_w, patch_x:patch_x + tile_w, c, n] += \
                        input[:, :, c, tile_ind, n, off_ind]

    err = np.reshape(output_d.asarray(), output.shape, order = 'F') - output
    assert np.max(np.abs(err)) < 10**-4, "Error in cudamatextras.tiled_to_flat_no_overlap"


def test_dot_block():
    block_size1 = 3
    block_size2 = 2
    num_blocks = 3
    dims = 4
    # CASE A: False False
    pm1a = np.array(np.random.randn(dims,block_size1*num_blocks), dtype=np.float32, order='F')
    pm2a = np.array(np.random.randn(block_size1,block_size2*num_blocks), dtype=np.float32, order='F')
    resa = np.array(np.empty((dims,block_size2*num_blocks)), dtype=np.float32, order='F')
    cm1a = cm.CUDAMatrix(pm1a)
    cm2a = cm.CUDAMatrix(pm2a)
    cresa = cm.CUDAMatrix(resa)
    for nn in range(num_blocks):
        resa[:,nn*block_size2:(nn+1)*block_size2] = np.dot(pm1a[:,nn*block_size1:(nn+1)*block_size1],pm2a[:,nn*block_size2:(nn+1)*block_size2]).copy()
        
    cme.dot_block(cm1a, block_size1, False, cm2a, block_size2, False, cresa)

    cresa.copy_to_host()
    
    # CASE B: True False
    pm1b = np.array(np.random.randn(dims,block_size1*num_blocks), dtype=np.float32, order='F')
    pm2b = np.array(np.random.randn(dims,block_size2*num_blocks), dtype=np.float32, order='F')
    resb = np.array(np.empty((block_size1,block_size2*num_blocks)), dtype=np.float32, order='F')
    cm1b = cm.CUDAMatrix(pm1b)
    cm2b = cm.CUDAMatrix(pm2b)
    cresb = cm.CUDAMatrix(resb)
    for nn in range(num_blocks):
        resb[:,nn*block_size2:(nn+1)*block_size2] = np.dot(pm1b[:,nn*block_size1:(nn+1)*block_size1].T,pm2b[:,nn*block_size2:(nn+1)*block_size2]).copy()
        
    cme.dot_block(cm1b, block_size1, True, cm2b, block_size2, False, cresb)

    cresb.copy_to_host()
    
    # CASE C: False True
    pm1c = np.array(np.random.randn(dims,block_size2*num_blocks), dtype=np.float32, order='F')
    pm2c = np.array(np.random.randn(block_size1,block_size2*num_blocks), dtype=np.float32, order='F')
    resc = np.array(np.empty((dims,block_size1*num_blocks)), dtype=np.float32, order='F')
    cm1c = cm.CUDAMatrix(pm1c)
    cm2c = cm.CUDAMatrix(pm2c)
    cresc = cm.CUDAMatrix(resc)
    for nn in range(num_blocks):
        resc[:,nn*block_size1:(nn+1)*block_size1] = np.dot(pm1c[:,nn*block_size2:(nn+1)*block_size2],pm2c[:,nn*block_size2:(nn+1)*block_size2].T).copy()
        
    cme.dot_block(cm1c, block_size2, False, cm2c, block_size2, True, cresc)

    cresc.copy_to_host()

    assert np.max(np.abs(resa - cresa.numpy_array)) + np.max(np.abs(resb - cresb.numpy_array))  + np.max(np.abs(resc - cresc.numpy_array)) < 10**-4, "Error in CUDAMatrix.dot_block exceeded threshold"

if __name__ == '__main__':
    nose.run()
