# training Tiled mPoT on high-resolution natural images 
# using Fast Persistent Contrastive Divergence. 
# Samples are drawn from the model using Hamiltonian Monte Carlo.
# Each input image is pre-processed by: 1) subtracting off the image mean intensity
# from all pixels and 2) dividing all pixels by the image standard deviation.
# 
# For simplicity of notation, let's consider a single image patch x.
# The free energy is: 
# F(x) = sum_j gamma log(1 + .5 sum_f P_fj (sum_i C_fi x_i)^2) +
#        + .5 sum_i x_i^2 - sum_k log(1+exp(sum_i W_ki x_i +b_k)) 
# Probability distribution: p(x) proportional to exp(-F(x))
# 
# This corresponds to an energy function defined over two sets of latent variables.
# The latent variables associated to the first term (the covariance latent variables) are
# gamma distributed, while those associated with the last term (the mean latent variables)
# are Bernoulli distributed.
#
# In the code below, we call:
#  gamma -> scale
#  C -> VF
#  W -> w_mean
#  b -> bias_mean
#
# In order to extend this model to high-resolution images we divide paramters into sets.
# Each set is used to tile the image without overlap. Filters in different sets do overlap 
# with a fixed diagonal offset (to compensate for each other's tiling artifacts). 
#
# For more information aobut the model, see:
# M. Ranzato, V. Mnih, G. Hinton, "Generating More Realistic Images Using Gated MRF's". 
# Advances in Neural Information Processing Systems 23 (NIPS 2010), Vancouver, 2010
# available at http://www.cs.toronto.edu/~ranzato/publications/ranzato_nips2010.pdf
#
# Marc'Aurelio Ranzato
# 23 December 2010
# ranzato@cs.toronto.edu

import sys
import numpy as np
import cudamat as cmt
import cudamatextras as cme
from scipy.io import loadmat, savemat
import gpu_lock
from ConfigParser import *
import os

def compute_energy_TmPoT(rawdata,tiledata,data,vel,energy,VF,scale,t13,t14,feat,featsq,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean, tt0,tt1,tt2):
    ''' compute the value of the hamiltonian (sum of model free energy and kinetic energy) over the whole input image
    this returns a scalar value for each input image in the batch'''
    # extract patches from input image
    cme.flat_to_tiled_diag(rawdata, tiledata, num_channels, num_replicas, stride, num_loc)
    # "mean" contribution to free energy
    cme.dot_block(tiledata, num_replicas**2 * batch_size, True, w_mean, num_filt_loc_mean, False, feat_mean)
    feat_mean.add_row_vec(bias_mean) 
    feat_mean.greater_than(tt0,target = tt1) # avoid numerical overflow while computing log(1+exp())
    tt1.mult(-1.,target=tt2)
    tt2.add(1.)    
    tt1.mult(feat_mean)
    feat_mean.mult(tt2)
    cmt.exp(feat_mean) 
    feat_mean.add(1.)
    cmt.log(feat_mean)
    feat_mean.mult(tt2)
    feat_mean.add(tt1)
    feat_mean.mult(-1.)
    feat_mean.sum(axis=1, target = t14)
    t14.reshape((num_replicas**2,batch_size))
    t14.sum(axis=0, target = energy)
    t14.reshape((num_replicas**2 * batch_size,1))
    # "covariance" contribution to free energy
    cmt.dot(pcatransf, tiledata, target = data) # covariance filters have whitening front-end
    cme.dot_block(data, num_replicas**2 * batch_size, True, VF, num_filt_loc_cov, False, feat)
    feat.mult(feat, target = featsq)   
    featsq.mult(0.5)
    featsq.add(1.)
    cmt.log(featsq)
    featsq.mult_by_row(scale) 
    featsq.sum(axis=1,target = t14)
    t14.reshape((num_replicas**2,batch_size))
    energy.add_sums(t14,axis=0)
    t14.reshape((num_replicas**2 * batch_size,1))
    # quadratic term
    rawdata.mult(rawdata, target = t13)
    energy.add_sums(t13, axis = 0,mult=0.5)
    # kinetic term
    vel.mult(vel, target = t13)
    energy.add_sums(t13, axis = 0, mult = .5)

def compute_gradient_TmPoT(rawdata,tiledata,data,VF,scale,feat,featsq,gradient,tilegradient,rawgradient,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean):
    ''' derivatives of the free energy w.r.t. input'''
    # extract patches from input images
    cme.flat_to_tiled_diag(rawdata, tiledata, num_channels, num_replicas, stride, num_loc)
    # "mean" contribution to gradient
    cme.dot_block(tiledata, num_replicas**2 * batch_size, True, w_mean, num_filt_loc_mean, False, feat_mean)
    feat_mean.add_row_vec(bias_mean) 
    feat_mean.apply_sigmoid() 
    cme.dot_block(w_mean, num_filt_loc_mean, False, feat_mean, num_filt_loc_mean, True, tilegradient)
    tilegradient.mult(-1.)
    # "covariance" contribution to gradient
    cmt.dot(pcatransf, tiledata, target = data) 
    cme.dot_block(data, num_replicas**2 * batch_size, True, VF, num_filt_loc_cov, False, feat)
    feat.mult(feat, target = featsq)
    featsq.mult(0.5)
    featsq.add(1.)
    featsq.reciprocal()
    featsq.mult_by_row(scale)
    feat.mult(featsq)
    cme.dot_block(VF, num_filt_loc_cov, False, feat, num_filt_loc_cov, True, gradient)
    tilegradient.add_dot(pcatransf.T,gradient)
    # accumulate gradients from overlapping patches into input image
    cme.tiled_to_flat_diag(tilegradient, rawgradient, num_channels, num_replicas, stride, num_loc)
    # gradient of quadratic term
    rawgradient.add(rawdata)

def draw_HMC_samples(rawdata,negtiledata,rawnegdata,negdata,vel,gradient,tilegradient,rawgradient,new_energy,old_energy,VF,scale,hmc_step,hmc_step_nr,hmc_ave_rej,hmc_target_ave_rej,t5,t7,t13,t14,t15,thresh,feat,featsq,batch_size,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,mask,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean, tt0,tt1,tt2):
    ''' Hamiltonian Monte Carlo sampler (just one step composed of many leap-frog steps)'''
    vel.fill_with_randn()
    vel.mult_by_col(mask) # condition on the borders
    rawnegdata.assign(rawdata)
    compute_energy_TmPoT(rawnegdata,negtiledata,negdata,vel,old_energy,VF,scale,t13,t14,feat,featsq,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean, tt0,tt1,tt2)
    compute_gradient_TmPoT(rawnegdata,negtiledata,negdata,VF,scale,feat,featsq,gradient,tilegradient,rawgradient,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean)
    rawgradient.mult_by_col(mask) # condition on the borders
    # half step
    vel.add_mult(rawgradient, -0.5*hmc_step)
    rawnegdata.add_mult(vel,hmc_step)
    # full leap-frog steps
    for ss in range(hmc_step_nr - 1):
        ## re-evaluate the gradient
        compute_gradient_TmPoT(rawnegdata,negtiledata,negdata,VF,scale,feat,featsq,gradient,tilegradient,rawgradient,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean)
        rawgradient.mult_by_col(mask) # condition on the borders
        # update variables
        vel.add_mult(rawgradient, -hmc_step)
        rawnegdata.add_mult(vel,hmc_step)
    # final half-step
    compute_gradient_TmPoT(rawnegdata,negtiledata,negdata,VF,scale,feat,featsq,gradient,tilegradient,rawgradient,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean)
    rawgradient.mult_by_col(mask) # condition on the borders
    vel.add_mult(rawgradient, -0.5*hmc_step)
    # compute new value of energy
    compute_energy_TmPoT(rawnegdata,negtiledata,negdata,vel,new_energy,VF,scale,t13,t14,feat,featsq,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,batch_size,filt_size,w_mean,bias_mean,num_filt_loc_mean, feat_mean, tt0,tt1,tt2)
    # rejecton
    old_energy.subtract(new_energy, target = thresh)
    cmt.exp(thresh)
    t15.fill_with_rand()
    t15.greater_than(thresh) # 1's detect rejections
    t15.sum(axis = 1, target = t5)
    t5.copy_to_host()
    rej = t5.numpy_array[0,0]/batch_size
    rawdata.mult_by_row(t15, target = t13)
    rawnegdata.mult_by_row(t15, target = t7)
    rawnegdata.subtract(t7)
    rawnegdata.add(t13)
    hmc_ave_rej = 0.9*hmc_ave_rej + 0.1*rej # running average rejection rate
    if hmc_ave_rej < hmc_target_ave_rej:
        hmc_step = min(hmc_step*1.01,0.25)
    else:
        hmc_step = max(hmc_step*0.99,.001)
    return hmc_step, hmc_ave_rej

def train_TmPoT():
    ''' Tiled mPoT trainer: sweeps over the training set to update the parameters.
    For each batch of samples compute derivatives w.r.t. parameters
    at the training samples and at the negative samples drawn by HMC sampler.
    Training using Fast Persistent Contrastive Divergence.'''

    config = ConfigParser()
    config.read('input_configuration')
    
    num_epochs = config.getint('MAIN_PARAMETER_SETTING','num_epochs')
    batch_size = 32
    startwd = config.getint('MAIN_PARAMETER_SETTING','startwd')
    
    # model parameters
    filt_size = config.getint('MODEL_PARAMETER_SETTING','filt_size') 
    stride = config.getint('MODEL_PARAMETER_SETTING','stride')
    num_filt_loc_cov = config.getint('MODEL_PARAMETER_SETTING','num_filt_loc_cov')
    num_replicas = config.getint('MODEL_PARAMETER_SETTING','num_replicas')
    num_loc = int(filt_size/stride) # number of filter sets
    side_image = num_replicas*filt_size+stride*(num_loc-1) # side of actual input image
    num_fac = num_loc * num_filt_loc_cov # (sqrt of) number of filter responses
    num_hid_cov = num_fac # (sqrt of) number of "covariance" latent variables 
    num_channels = 1 # number of input channels
    num_vis = side_image**2 * num_channels # number of input units
    num_filt_loc_mean = config.getint('MODEL_PARAMETER_SETTING','num_filt_loc_mean')
    num_hid_mean = num_loc * num_filt_loc_mean # (sqrt of) number of "mean" latent variables
    
    # load data
    data_file_name =  config.get('DATA','data_file_name')
    pca_file_name =  config.get('DATA','pca_file_name')
    dd = loadmat(data_file_name)
    nrima = dd["data"].shape[0]
    side = np.sqrt(dd["data"].shape[1]/num_channels) # >= side_image
    dat = np.reshape(dd["data"],(nrima,side,side),order='F') # need to crop large patches
    pca = loadmat(pca_file_name)
    num_visPCA = pca["pcatransf"].shape[0]
    num_batches = nrima    
    
    # training parameters
    epsilon = config.getfloat('OPTIMIZER_PARAMETERS','epsilon')
    epsilonVF = 1.*epsilon # learning rates
    epsilonb = 0.01*epsilon
    epsilonw_mean = 0.1*epsilon
    epsilonb_mean = 0.1*epsilon
    epsilonwg_mean = 0.01*epsilon
    epsilonbg_mean = 0.01*epsilon
    weightcost_final =  config.getfloat('OPTIMIZER_PARAMETERS','weight_decay')

    # HMC setting
    hmc_step_nr = 20
    hmc_step =  .01
    hmc_ave_rej =  .1
    hmc_target_ave_rej = .1

    back_up = config.getint('BACK_UP','back_up')
    back_up_freq = config.getint('BACK_UP','back_up_freq')
    save_dir =  config.get('BACK_UP','save_dir')
    exper_id = "parameters"

    # initialize parameters at random
    VF = cmt.CUDAMatrix(0.02 * np.random.randn(num_visPCA, num_fac)) 
    scale = cmt.CUDAMatrix(0.4*np.ones((1,num_hid_cov)))
    w_mean = cmt.CUDAMatrix(0.01*np.random.randn(num_channels*filt_size**2, num_hid_mean)) 
    bias_mean = cmt.CUDAMatrix(-1.0*np.ones((1,num_hid_mean)))
    rawnegdata = cmt.CUDAMatrix(np.zeros((num_vis, batch_size))) 
    normVF = 1
    # auxiliary parameters used by Fast PCD initialized at 0
    VFf = cmt.CUDAMatrix(np.zeros((num_visPCA, num_fac))) 
    w_meanf = cmt.CUDAMatrix(np.zeros((num_channels*filt_size**2, num_hid_mean)) ) 
    bias_meanf = cmt.CUDAMatrix(np.zeros((1,num_hid_mean)) )    
    VFt = cmt.CUDAMatrix(np.zeros((num_visPCA, num_fac))) 
    w_meant = cmt.CUDAMatrix(np.zeros((num_channels*filt_size**2, num_hid_mean)) ) 
    bias_meant = cmt.CUDAMatrix(np.zeros((1,num_hid_mean)) )
    
    # initialize weight updates
    VFinc = cmt.CUDAMatrix(np.zeros((num_visPCA, num_fac)))
    scaleinc = cmt.CUDAMatrix(np.zeros((1,num_hid_cov)))
    w_meaninc = cmt.CUDAMatrix(np.zeros((num_channels*filt_size**2, num_hid_mean)))
    bias_meaninc = cmt.CUDAMatrix(np.zeros((1,num_hid_mean)))
    # initialize other variables
    rawdata = cmt.CUDAMatrix(np.empty((num_vis, batch_size))) 
    tiledata = cmt.CUDAMatrix(np.empty((num_channels*filt_size**2, num_replicas**2 * batch_size * num_loc ))) 
    data = cmt.CUDAMatrix(np.empty((num_visPCA, num_replicas**2 * batch_size * num_loc))) 
    negdataini = cmt.CUDAMatrix(np.empty((num_vis, batch_size))) 

    feat = cmt.CUDAMatrix(np.empty((num_replicas**2 * batch_size, num_fac)))
    featsq = cmt.CUDAMatrix(np.empty((num_replicas**2 * batch_size, num_fac)))
    feat_mean = cmt.CUDAMatrix(np.empty((num_replicas**2 * batch_size, num_hid_mean)))

    rawgradient = cmt.CUDAMatrix(np.empty((num_vis, batch_size))) 
    tilegradient = cmt.CUDAMatrix(np.empty((num_channels*filt_size**2, num_replicas**2 * batch_size * num_loc))) 
    gradient = cmt.CUDAMatrix(np.empty((num_visPCA, num_replicas**2 * batch_size * num_loc))) 

    thresh = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    old_energy = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    new_energy = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    vel = cmt.CUDAMatrix(np.random.randn(num_vis, batch_size))
    pcatransf = cmt.CUDAMatrix(pca["pcatransf"]) 
    # define conditioning mask
    maskpy = np.zeros((side_image,side_image))
    maskpy[stride*(num_loc-1): stride*(num_loc-1) + (num_replicas-1)*filt_size,stride*(num_loc-1): stride*(num_loc-1) + (num_replicas-1)*filt_size] = 1.
    mask = cmt.CUDAMatrix(np.reshape(maskpy,(num_vis,1)))
    
    # temporary vars
    t1 = cmt.CUDAMatrix(np.empty((num_replicas**2 * batch_size, num_hid_cov)))
    t2 = cmt.CUDAMatrix(np.zeros((num_channels*filt_size**2, num_hid_mean)))
    t5 = cmt.CUDAMatrix(np.empty((1,1)))
    t7 = cmt.CUDAMatrix(np.empty((num_vis, batch_size)))
    t8 = cmt.CUDAMatrix(np.empty((num_visPCA, num_fac)))
    t11 = cmt.CUDAMatrix(np.empty((1,num_fac)))
    t13 = cmt.CUDAMatrix(np.empty((num_vis, batch_size)))
    t14 = cmt.CUDAMatrix(np.ones((num_replicas**2 *  batch_size,1)))
    t15 = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    t20 = cmt.CUDAMatrix(1./num_hid_cov*np.ones((1,num_hid_cov)))
    tt0 = cmt.CUDAMatrix(15.*np.ones((num_replicas**2 * batch_size, num_hid_mean)))
    tt1 = cmt.CUDAMatrix(np.empty((num_replicas**2 * batch_size, num_hid_mean)))
    tt2 = cmt.CUDAMatrix(np.empty((num_replicas**2 * batch_size, num_hid_mean)))

    # start training
    for epoch in range(num_epochs):

        print "Epoch " + str(epoch + 1)
        
        # anneal learning rates (after a bit)
        if epoch <= 100: 
            epsilonVFc    = epsilonVF
            epsilonbc    = epsilonb
            epsilonw_meanc = epsilonw_mean
            epsilonb_meanc = epsilonb_mean
            epsilonwg_meanc = epsilonwg_mean
            epsilonbg_meanc = epsilonbg_mean
        else:
            epsilonVFc    = epsilonVF/max(1.,epoch/200.)
            epsilonbc    = epsilonb/max(1.,epoch/200.)
            epsilonw_meanc = epsilonw_mean/max(1.,epoch/200.)
            epsilonb_meanc = epsilonb_mean/max(1.,epoch/200.)
            epsilonwg_meanc = epsilonwg_mean/max(1.,epoch/200.)
            epsilonbg_meanc = epsilonbg_mean/max(1.,epoch/200.)
        weightcost = weightcost_final
        
        if epoch <= startwd:	
            weightcost = 0
        
        for batch in range(num_batches):
            # crop large patches from images at random locations
            rrr = int(np.random.rand()*(side-side_image))
            ccc = int(np.random.rand()*(side-side_image))
            nnn = int(np.random.rand()*(nrima - batch_size))
            rawdata = cmt.CUDAMatrix(np.reshape(dat[np.mod(range(nnn,nnn+batch_size),nrima),rrr:rrr+side_image,ccc:ccc+side_image],(batch_size,side_image**2 * num_channels),order='F').T)

            # extract patches from input images
            cme.flat_to_tiled_diag(rawdata, tiledata, num_channels, num_replicas, stride, num_loc)
            # "mean" part
            cme.dot_block(tiledata, num_replicas**2 * batch_size, True, w_mean, num_filt_loc_mean, False, feat_mean)
            feat_mean.add_row_vec(bias_mean) 
            feat_mean.apply_sigmoid() 
            feat_mean.mult(-1.)
            cme.dot_block(tiledata, num_replicas**2 * batch_size, False, feat_mean, num_filt_loc_mean, False, w_meaninc)
            feat_mean.sum(axis=0,target = bias_meaninc)
            # "covariance" part
            cmt.dot(pcatransf, tiledata, target = data) 
            cme.dot_block(data, num_replicas**2 * batch_size, True, VF, num_filt_loc_cov, False, feat)
            feat.mult(feat, target = featsq)
            featsq.mult(0.5)
            featsq.add(1.)
            featsq.reciprocal(target = t1)
            t1.mult_by_row(scale)
            feat.mult(t1)
            cme.dot_block(data, num_replicas**2 * batch_size, False, feat, num_filt_loc_cov, False, VFinc)
            cmt.log(featsq)
            featsq.sum(axis=0,target = scaleinc)

            # update total weight vector for Fast PCD
            VF.add(VFf,target = VFt)
            w_mean.add(w_meanf, target = w_meant)
            bias_mean.add(bias_meanf, target = bias_meant)

            # HMC sampling
            negdataini.assign(rawnegdata)
            hmc_step, hmc_ave_rej = draw_HMC_samples(negdataini,tiledata,rawnegdata,data,vel,gradient,tilegradient,rawgradient,new_energy,old_energy,VFt,scale,hmc_step,hmc_step_nr,hmc_ave_rej,hmc_target_ave_rej,t5,t7,t13,t14,t15,thresh,feat,featsq,batch_size,pcatransf,num_loc,num_filt_loc_cov,stride,num_replicas,num_channels,mask,filt_size,w_meant,bias_meant,num_filt_loc_mean, feat_mean, tt0,tt1,tt2)

            # extract patches from negative particles
            cme.flat_to_tiled_diag(rawnegdata, tiledata, num_channels, num_replicas, stride, num_loc)
            # "mean" part
            cme.dot_block(tiledata, num_replicas**2 * batch_size, True, w_meant, num_filt_loc_mean, False, feat_mean)
            feat_mean.add_row_vec(bias_meant) 
            feat_mean.apply_sigmoid() 
            cme.dot_block(tiledata, num_replicas**2 * batch_size, False, feat_mean, num_filt_loc_mean, False, t2)
            w_meaninc.add(t2)
            bias_meaninc.add_sums(feat_mean, axis = 0)
            # "covariance" part
            cmt.dot(pcatransf, tiledata, target = data) 
            cme.dot_block(data, num_replicas**2 * batch_size, True, VFt, num_filt_loc_cov, False, feat)
            feat.mult(feat, target = featsq)
            featsq.mult(0.5)
            featsq.add(1.)
            featsq.reciprocal(target = t1)
            t1.mult_by_row(scale)
            feat.mult(t1)
            cme.dot_block(data, num_replicas**2 * batch_size, False, feat, num_filt_loc_cov, False, t8)
            VFinc.subtract(t8)
            cmt.log(featsq)
            featsq.mult(-1.)
            scaleinc.add_sums(featsq, axis = 0)
            # average gradients of <scale>
            scaleinc.sum(axis = 1, target = t5)
            t5.copy_to_host()
            t20.mult(float(t5.numpy_array[0,0]), target = scaleinc)

            # update parameters
            # fast parameters
            w_meanf.mult(19./20)
            w_meanf.add_mult(w_meaninc, -epsilonw_mean/(num_replicas**2 * batch_size))
            bias_meanf.mult(19./20)
            bias_meanf.add_mult(bias_meaninc, -epsilonb_mean/(num_replicas**2 * batch_size))
            VFf.mult(19./20)
            VFf.add_mult(VFinc, -epsilonVF/(num_replicas**2 * batch_size))
            # actual parameters
            w_meaninc.add_mult(w_mean.sign(),weightcost)
            w_mean.add_mult(w_meaninc, -epsilonw_meanc/(num_replicas**2 * batch_size))
            bias_mean.add_mult(bias_meaninc, -epsilonb_meanc/(num_replicas**2 * batch_size))
            VFinc.add_mult(VF.sign(), weightcost)
            VF.add_mult(VFinc, -epsilonVFc/(num_replicas**2 * batch_size))
            # normalize cols of VF
            VF.mult(VF, target = t8)
            t8.sum(axis = 0, target = t11)
            cmt.sqrt(t11)
            t11.sum(axis=1,target = t5)
            t5.copy_to_host()
            normVF = .95*normVF + (.05/num_fac) * t5.numpy_array[0,0] # running estimate of norm
            t11.reciprocal()
            VF.mult_by_row(t11) 
            VF.mult(normVF)
            scale.add_mult(scaleinc, -epsilonbc/(num_replicas**2 * batch_size))

        # back-up
        print "VF: " + '%3.2e' % VF.euclid_norm() + ", DVF: " + '%3.2e' % (VFinc.euclid_norm()*(epsilonVFc/(num_replicas**2 * batch_size ))) + ", scale: " + '%3.2e' % scale.euclid_norm() + ", Dscale: " + '%3.2e' % (scaleinc.euclid_norm()*(epsilonbc/(num_replicas**2 * batch_size ))) + ", wm: " + '%3.2e' % w_mean.euclid_norm() + ", Dwm: " + '%3.2e' % (w_meaninc.euclid_norm()*(epsilonw_meanc/(num_replicas**2 * batch_size ))) + ", bm: " + '%3.2e' % bias_mean.euclid_norm() + ", Dbm: " + '%3.2e' % (bias_meaninc.euclid_norm()*(epsilonb_meanc/(num_replicas**2 * batch_size ))) + ", step: " + '%3.2e' % hmc_step  +  ", rej: " + '%3.2e' % hmc_ave_rej 
        print "Fast VF: " + '%3.2e' % VFf.euclid_norm() + ", wm: " + '%3.2e' % w_meanf.euclid_norm() + ", bm: " + '%3.2e' % bias_meanf.euclid_norm() 
        sys.stdout.flush()

        if np.mod(epoch,back_up_freq) == 0:
            VF.copy_to_host()
            scale.copy_to_host()
            w_mean.copy_to_host()
            bias_mean.copy_to_host()
            rawdata.copy_to_host()
            rawnegdata.copy_to_host()
            savemat(save_dir + "temp_params", {'VF':VF.numpy_array,'scale': scale.numpy_array, 'w_mean': w_mean.numpy_array, 'bias_mean': bias_mean.numpy_array, 'epoch':epoch, 'rawdata':rawdata.numpy_array, 'rawnegdata':rawnegdata.numpy_array,'exper_id':exper_id})
        
        if np.mod(epoch,100) == 0:
            sys.stdout.flush()
            VF.copy_to_host()
            scale.copy_to_host()
            w_mean.copy_to_host()
            bias_mean.copy_to_host()
            rawdata.copy_to_host()
            rawnegdata.copy_to_host()
            savemat(save_dir + exper_id + "_" + str(epoch), {'VF':VF.numpy_array,'scale': scale.numpy_array, 'w_mean': w_mean.numpy_array, 'bias_mean': bias_mean.numpy_array,'epoch':epoch, 'rawdata':rawdata.numpy_array, 'rawnegdata':rawnegdata.numpy_array,'exper_id':exper_id})
       
    VF.copy_to_host()
    scale.copy_to_host()
    w_mean.copy_to_host()
    bias_mean.copy_to_host()
    rawdata.copy_to_host()
    rawnegdata.copy_to_host()
    savemat(save_dir + exper_id, {'VF':VF.numpy_array,'scale': scale.numpy_array, 'w_mean': w_mean.numpy_array, 'bias_mean': bias_mean.numpy_array, 'epoch':epoch, 'rawdata':rawdata.numpy_array, 'rawnegdata':rawnegdata.numpy_array,'exper_id':exper_id})
    
if __name__ == "__main__":
  # initialize CUDA
  cmt.cuda_set_device(gpu_lock.obtain_lock_id())
  cmt.cublas_init()
  cmt.CUDAMatrix.init_random(1)
  train_TmPoT()
  cmt.cublas_shutdown()











